import os, torch
import numpy as np
import pandas as pd
from torch import nn
import torch.nn.functional as F
from sklearn.preprocessing import MinMaxScaler
import matplotlib.pyplot as plt
import argparse
from PROMINENT_methylation.train import *

def preprocessing1(path_fn, gene_fn, label_fn):
    ### pathway datasets
    if os.path.exists(path_fn):
        pathway = pd.read_csv(path_fn, header=0)       
        print(">> Pathway Data :",path_fn)
        pathway_info = pathway.iloc[:,1:]
        pathway_info = pathway_info.values
        pathway_info = np.transpose (pathway_info)
        pathway_info = torch.FloatTensor(pathway_info)
        print("pathway matrix shape : ",pathway_info.shape)
        print("num_pathway : ",pathway_info.shape[0])
    else:
        pathway_info = None
        
    ### methylation datasets
    print(">> Methylation Gene-level Data:",gene_fn)
    data = pd.read_csv(gene_fn, header=0)

    expression = data.iloc[:,1:]
    features = data.iloc[:,0].tolist()
    expression = expression.values
    expression = np.transpose(expression)

    scaler = MinMaxScaler()
    scaler = scaler.fit(expression)
    expression = scaler.transform(expression)

    sample_dim = expression.shape[0]
    input_dim = expression.shape[1]

    #print dimension of sample and number of genes
    print("sample_dim : ",sample_dim)
    print("input_size (number of genes): ",input_dim)
    
    pheno = pd.read_csv(label_fn, index_col=0)
    status = np.array(pheno['label']).reshape(-1,1)

    patient = list(data.iloc[:,1:].columns.values.tolist()) 
    print("patient list : ",patient[0:6])
    print("feature list : ",features[0:6])
    
    return pathway_info, expression, status, features


def train():
    parser = argparse.ArgumentParser(description='5 fold CV')
    parser.add_argument('--input_csv', default='gene.average.beta.by.intensity.csv', help='Input csv file of X_i. Rows are genes, columns are samples.')
    parser.add_argument('--input_label', default='label.csv', help='Input label file path.')
    parser.add_argument('--input_path', default='pathway_gobp.csv', help='Pathway information from dataprep.')
    parser.add_argument('--output', default='pred.pkl', help='Prediction output file path. pkl file.')
    parser.add_argument('--output_shap', default='shap.pkl', help='Model interpretation output file path. pkl file.')
    parser.add_argument('--mlp', action='store_true', help='DNN only model.')

    args = parser.parse_args()

    input_file_csv = args.input_csv
    input_file_label = args.input_label
    input_file_path = args.input_path
    output_file = args.output
    output_file_shap = args.output_shap
    mlp = args.mlp
    
    pathway_info, x_data, status, features = preprocessing1(input_file_path, input_file_csv,input_file_label)
    
    trainArgs = {}
    trainArgs['x_data'] = x_data
    trainArgs['y_data'] = status
    trainArgs['pathway_info'] = pathway_info
    trainArgs['features'] = features
    trainArgs['num_fc_list'] = [32]
    trainArgs['lr_list'] = [0.001]
    trainArgs['device'] = '0'
    trainArgs['seed'] = 0
    trainArgs['filename'] = output_file
    trainArgs['filename2'] = output_file_shap
    
    if mlp:
        print("Run PROMINENT_DNN.")
        train = train_kfold_mlp(trainArgs)
    else:
        train = train_kfold2(trainArgs)
        
    result = train.kfold()
    
if __name__ == '__main__':
    train()
    



import torch, os
import pandas as pd
import numpy as np
from torch import nn
from torch.utils.data import Dataset

class EarlyStopping:
    def __init__(self, patience=20, verbose=False, delta=0, path='checkpoint.pt'):
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_auc_min = 0
        self.delta = delta
        self.path = path

    def __call__(self, val_auc, model, epoch):

        score = val_auc
        if epoch > 0 :
            if self.best_score is None:
                self.best_score = score
                self.save_checkpoint(score, model)
            elif score > self.best_score + self.delta:
                self.best_score = score
                self.save_checkpoint(score, model)
                self.counter = 0
            else:
                self.counter += 1
                if self.counter >= self.patience:
                    self.early_stop = True

    def save_checkpoint(self, val_auc, model):
        if self.verbose:
            print(f'Validation AUC increased ({self.val_auc_min:.6f} --> {val_auc:.6f}).  Saving model ...')
        torch.save(model, self.path)
        self.val_auc_min = val_auc
                
class CustomDataset(Dataset): 
    def __init__(self,x,y):
        self.x_data = x
        self.y_data = y

    def __len__(self): 
        return len(self.x_data)

    def __getitem__(self, idx): 
        x = torch.FloatTensor(self.x_data[idx])
        y = torch.LongTensor(self.y_data[idx])
        return x, y


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import shap
import pickle
import seaborn as sns
import argparse

def get_feature_importance():
    parser = argparse.ArgumentParser(description='Caculate the importance for each feature and pathway.')
    parser.add_argument('--input_shap', default='shap.pkl', help='SHAP output from training.')
    parser.add_argument('--output_feature', default='Feature_importance.csv', help='Output feature importance file path. csv file.')
    parser.add_argument('--output_pathway', default='Pathway_importance.csv', help='Output pathway importance file path. csv file.')
    args = parser.parse_args()

    input_file_shap = args.input_shap
    output_file_feat = args.output_feature
    output_file_path = args.output_pathway
    
    with open(input_file_shap, 'rb') as file:
        shap_dict = pickle.load(file)
    shaps = shap_dict["path"]
    shaps_feat = shap_dict["feat"]
    features = pd.read_csv("pathway_gobp.csv",index_col=0).columns.tolist()
    result_array = np.zeros_like(np.sum(np.abs(shaps[0][1]), axis=0)[32:])
    pos_arr = np.zeros_like(np.sum(np.abs(shaps[0][1]), axis=0)[32:])
    neg_arr = np.zeros_like(np.sum(np.abs(shaps[0][1]), axis=0)[32:])
    for i, shap in enumerate(shaps):
        feature_importance = np.sum(np.abs(shap[1]), axis=0)[32:]
        positive_sum = np.sum(shap[1] * (shap[1] > 0), axis=0)[32:]
        negative_sum = np.sum(shap[1] * (shap[1] < 0), axis=0)[32:]
        result_array += feature_importance
        pos_arr += positive_sum
        neg_arr += negative_sum
    fi_result = pd.DataFrame({"path_name":features,"importance":result_array, "pos":pos_arr,"neg": np.abs(neg_arr)})
    df_sorted = fi_result.sort_values(by='importance', ascending=False)
    df_sorted.to_csv(output_file_path, index=False)

    features = pd.read_csv("gene.average.beta.by.intensity.csv",index_col=0).index.tolist()
    result_array = np.zeros_like(np.sum(np.abs(shaps_feat[0][1]), axis=0))
    pos_arr = np.zeros_like(np.sum(np.abs(shaps_feat[0][1]), axis=0))
    neg_arr = np.zeros_like(np.sum(np.abs(shaps_feat[0][1]), axis=0))
    for i, shap in enumerate(shaps_feat):
        feature_importance = np.sum(np.abs(shap[1]), axis=0)
        positive_sum = np.sum(shap[1] * (shap[1] > 0), axis=0)
        negative_sum = np.sum(shap[1] * (shap[1] < 0), axis=0)
        result_array += feature_importance
        pos_arr += positive_sum
        neg_arr += negative_sum
    fi_result = pd.DataFrame({"path_name":features,"importance":result_array, "pos":pos_arr,"neg": np.abs(neg_arr)})
    df_sorted = fi_result.sort_values(by='importance', ascending=False)
    df_sorted.to_csv(output_file_feat,index=False)
    
if __name__ == '__main__':
    get_feature_importance()


import argparse
import pandas as pd

def dataprep():
    parser = argparse.ArgumentParser(description='Prepare the data for training. It will generate the sparse matrix for masking.')
    parser.add_argument('--input_csv', default='gene.average.beta.by.intensity.csv', help='Input csv file of X_i. Rows are genes, columns are samples.')
    parser.add_argument('--input_gmt', default='c5.go.bp.v2023.1.Hs.symbols.gmt', help='Input pathway gene sets from MsigDB')
    parser.add_argument('--output', default='pathway_gobp.csv', help='Output file path. csv file.')

    args = parser.parse_args()

    input_file_csv = args.input_csv
    input_file_gmt = args.input_gmt
    output_file = args.output

    # read in X
    m = pd.read_csv(input_file_csv, index_col=0)
    m_genes = list(m.index)
    
    #read in gene sets and get the M
    path_dict = {}
    with open(input_file_gmt, 'r') as file:
        for line in file:
            columns = line.strip().split('\t')
            path = columns[0]
            path_genes = columns[2:]
            if len(list(set(m_genes) & set(path_genes))) >= 10:
                path_dict.update({path:list(set(m_genes) & set(path_genes))})
                
    df = pd.DataFrame(0, index=m_genes, columns = list(path_dict.keys()))
    for path, genes in path_dict.items():
        for gene in genes:
            df.loc[gene,path] = 1
    df = df.reset_index(drop=False)
    df.to_csv(output_file,index=None)

if __name__ == '__main__':
    dataprep()


import torch
import torch.nn as nn
    
def weights_init(m):
    if isinstance(m, nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight)
        torch.nn.init.zeros_(m.bias)    
        
class PINNet(nn.Module):
    def __init__(self,input_size, pathway, num_fc):
        super(PINNet, self).__init__()
        
        num_pw = pathway.shape[0]
        self.pathway = pathway

        self.fc1 = nn.Sequential(nn.Linear(input_size, num_fc),
                                 nn.BatchNorm1d(num_fc),
                                 nn.LayerNorm(num_fc),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc2 = nn.Sequential(nn.Linear(num_fc + num_pw, 64),
                                 nn.BatchNorm1d(64),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))      
        
        self.fc3 = nn.Sequential(nn.Linear(64,2),
                                 nn.Softmax(dim = 1))    
        
        self.pw_linear = nn.Linear(input_size, num_pw)
        self.norm_gene = torch.sqrt(pathway.sum(axis=1))
        self.pw = nn.Sequential(nn.BatchNorm1d(num_pw),
                                nn.LayerNorm(num_pw),
                                nn.Tanh(),
                                nn.Dropout(0.1))
                                
        self.fc1.apply(weights_init)   
        self.fc2.apply(weights_init)        
        self.fc3.apply(weights_init) 
        self.pw_linear.apply(weights_init) 

    def forward(self, input):     
        x1 = self.fc1(input)
        self.pw_linear.weight = nn.Parameter(self.pw_linear.weight * self.pathway)
        x2 = self.pw_linear(input)
        x2 = x2/self.norm_gene
        x2 = self.pw(x2)
        out = torch.cat((x1, x2), dim=1)
        out = self.fc2(out)
        out = self.fc3(out)
        return out
    
class PINNet2(nn.Module):
    def __init__(self,input_size, pathway, num_fc):
        super(PINNet2, self).__init__()
        
        num_pw = pathway.shape[0]
        self.pathway = pathway

        self.fc1 = nn.Sequential(nn.Linear(input_size, num_fc),
                                 nn.BatchNorm1d(num_fc),
                                 nn.LayerNorm(num_fc),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc2 = nn.Sequential(nn.Linear(num_pw, 128),
                                 nn.BatchNorm1d(128),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc3 = nn.Sequential(nn.Linear(128, 64),
                                 nn.BatchNorm1d(64),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc4 = nn.Sequential(nn.Linear(64,2),
                                 nn.Softmax(dim = 1))    
        
        self.pw_linear = nn.Linear(input_size, num_pw)
        self.norm_gene = torch.sqrt(pathway.sum(axis=1))
        self.pw = nn.Sequential(nn.BatchNorm1d(num_pw),
                                nn.LayerNorm(num_pw),
                                nn.Tanh(),
                                nn.Dropout(0.1))
                                
        self.fc1.apply(weights_init)   
        self.fc2.apply(weights_init)        
        self.fc3.apply(weights_init)
        self.fc4.apply(weights_init)
        self.pw_linear.apply(weights_init) 

    def forward(self, input):     
        self.pw_linear.weight = nn.Parameter(self.pw_linear.weight * self.pathway)
        x2 = self.pw_linear(input)
        x2 = x2/self.norm_gene
        x2 = self.pw(x2)
        out = self.fc2(x2)
        out = self.fc3(out)
        out = self.fc4(out)
        return out
    
class PINNet3(nn.Module):
    def __init__(self,input_size_gene, input_size_meth, pathway, num_fc):
        super(PINNet3, self).__init__()
        
        num_pw = pathway.shape[0]
        self.pathway = pathway

        self.fc1 = nn.Sequential(nn.Linear(input_size_meth, num_fc),
                                 nn.BatchNorm1d(num_fc),
                                 nn.LayerNorm(num_fc),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc2 = nn.Sequential(nn.Linear(num_fc + num_pw, 64),
                                 nn.BatchNorm1d(64),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))      
        
        self.fc3 = nn.Sequential(nn.Linear(64,2),
                                 nn.Softmax(dim = 1))    
        
        self.pw_linear = nn.Linear(input_size_gene, num_pw)
        self.norm_gene = torch.sqrt(pathway.sum(axis=1)+1)
        self.pw = nn.Sequential(nn.BatchNorm1d(num_pw),
                                nn.LayerNorm(num_pw),
                                nn.Tanh(),
                                nn.Dropout(0.1))
                                
        self.fc1.apply(weights_init)   
        self.fc2.apply(weights_init)        
        self.fc3.apply(weights_init) 
        self.pw_linear.apply(weights_init) 

    def forward(self, input_gene, input_meth):     
        x1 = self.fc1(input_meth)
        self.pw_linear.weight = nn.Parameter(self.pw_linear.weight * self.pathway)
        x2 = self.pw_linear(input_gene)
        x2 = x2/self.norm_gene
        x2 = self.pw(x2)
        out = torch.cat((x1, x2), dim=1)
        out = self.fc2(out)
        #print(out)
        #print(out.shape)
        out = self.fc3(out)
        #print(out)
        #print(out.shape)
        return out
    
class MLP(nn.Module):
    def __init__(self,input_size, pathway, num_fc):
        super(MLP, self).__init__()
        
        num_pw = pathway.shape[0]
        self.pathway = pathway

        self.fc1 = nn.Sequential(nn.Linear(input_size, num_fc),
                                 nn.BatchNorm1d(num_fc),
                                 nn.LayerNorm(num_fc),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))
        
        self.fc2 = nn.Sequential(nn.Linear(num_fc, 64),
                                 nn.BatchNorm1d(64),
                                 nn.Tanh(),
                                 nn.Dropout(0.1))      
        
        self.fc3 = nn.Sequential(nn.Linear(64,2),
                                 nn.Softmax(dim = 1))    
        
        self.pw_linear = nn.Linear(input_size, num_pw)
        self.norm_gene = torch.sqrt(pathway.sum(axis=1))
        self.pw = nn.Sequential(nn.BatchNorm1d(num_pw),
                                nn.LayerNorm(num_pw),
                                nn.Tanh(),
                                nn.Dropout(0.1))
                                
        self.fc1.apply(weights_init)   
        self.fc2.apply(weights_init)        
        self.fc3.apply(weights_init) 
        self.pw_linear.apply(weights_init) 

    def forward(self, input):     
        x1 = self.fc1(input)
        x2 = self.fc2(x1)
        out = self.fc3(x2)
        return out


import os, copy, torch, random, time, datetime
import numpy as np
import pandas as pd
from torch import nn
import torch.nn.functional as F
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from sklearn import metrics
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import train_test_split
from imblearn.over_sampling import SMOTE
import shap
import pickle
import matplotlib.pyplot as plt
import argparse
from sklearn.metrics import roc_curve, precision_recall_curve, auc
from sklearn.metrics import roc_auc_score, average_precision_score
from sklearn.metrics import precision_recall_curve, matthews_corrcoef,accuracy_score
from PROMINENT_methylation.model import *
from PROMINENT_methylation.utils import *

def preprocessing1(path_fn, gene_fn, label_fn):
    ### pathway datasets
    if os.path.exists(path_fn):
        pathway = pd.read_csv(path_fn, header=0)  
        path_features = list(pathway.columns[1:])
        print(">> Pathway Data :",path_fn)
        pathway_info = pathway.iloc[:,1:]
        pathway_info = pathway_info.values
        pathway_info = np.transpose (pathway_info)
        pathway_info = torch.FloatTensor(pathway_info)
        print("pathway matrix shape : ",pathway_info.shape)
        print("num_pathway : ",pathway_info.shape[0])
    else:
        print("No pathway information!")
        pathway_info = None
        
    ### methylation datasets
    print(">> Methylation Gene-level Data:",gene_fn)
    data = pd.read_csv(gene_fn, header=0)

    expression = data.iloc[:,1:]
    features = data.iloc[:,0].tolist()
    expression = expression.values
    expression = np.transpose(expression)

    scaler = MinMaxScaler()
    scaler = scaler.fit(expression)
    expression = scaler.transform(expression)

    sample_dim = expression.shape[0]
    input_dim = expression.shape[1]

    #print dimension of sample and number of genes
    print("sample_dim : ",sample_dim)
    print("input_size (number of genes): ",input_dim)
    
    pheno = pd.read_csv(label_fn, index_col=0)
    status = np.array(pheno['label']).reshape(-1,1)

    patient = list(data.iloc[:,1:].columns.values.tolist()) 
    print("patient list : ",patient[0:6])
    print("feature list : ",features[0:6])
    
    return pathway_info, expression, status, features,path_features

def softmax(x):
    # Subtract the maximum value from each element for numerical stability
    x_exp = np.exp(x - np.max(x, axis=1, keepdims=True))
    # Calculate the softmax probabilities
    return x_exp / np.sum(x_exp, axis=1, keepdims=True)
def fscore(true, prob, beta):
    precision,recall,_ = precision_recall_curve(true,prob)
    f = (1+beta**2)*(precision*recall)/(beta**2*precision+recall)
    idx = np.nanargmax(f)
    pr = precision[idx] 
    rc = recall[idx] 
    f = f[idx]
    return pr, rc, f
def calculate_metrics(y_true, y_pred):
    y_true = np.array(y_true).astype(int)
    TP = np.sum((y_pred == 1) & (y_true == 1))
    TN = np.sum((y_pred == 0) & (y_true == 0))
    FP = np.sum((y_pred == 1) & (y_true == 0))
    FN = np.sum((y_pred == 0) & (y_true == 1))
    TPR = TP / (TP + FN)
    TNR = TN / (TN + FP)
    PPV = TP / (TP + FP)
    NPV = TN / (TN + FN)
    return TPR, TNR, PPV, NPV
def get_threshold(y_true_train, y_prob_train):
    mccs = []
    thresholds = np.arange(0,1,0.001)
    for threshold in thresholds:
        y_pred_train = np.greater(y_prob_train, threshold).astype(int)
        mcc = matthews_corrcoef(y_true_train, y_pred_train)
        mccs.append(mcc)
    mccs = np.array(mccs)
    max_mcc = mccs.max() 
    max_mcc_threshold =  thresholds[mccs.argmax()]
    return max_mcc_threshold    

def mcc_score(y_true, y_prob,threshold):
    y_pred = np.greater(y_prob, threshold).astype(int)
    mcc = matthews_corrcoef(y_true, y_pred) 
    return mcc, threshold

class train_test:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : gene expression data
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        directory = './CV_best_model'
        if not os.path.exists(directory):
            os.makedirs(directory)

    def kfold(self):
        trainArgs = self.trainArgs
        expression1 = trainArgs['expression1']
        status1 = trainArgs['status1']
        expression = trainArgs['expression']
        status = trainArgs['status']
        input_dim =expression.shape[1]
        pathway_info = trainArgs['pathway_info'].to(self.device)
        num_pathway = pathway_info.shape[0]
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']
       
        
        kfold = StratifiedKFold(n_splits = 10, shuffle=True, random_state = random_seed)
        best_test_auc = 0
        best_fold = 0
        test_dataset1 = CustomDataset(expression1,status1) 
        test_loader1 = DataLoader(dataset = test_dataset1, batch_size = 64, shuffle = False)
        for fold, (train_index, test_index) in enumerate(kfold.split(expression, status)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = expression[train_index]
            y_train_ = status[train_index] 
            x_test = expression[test_index]  
            y_test = status[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            # smote = SMOTE(random_state=random_seed)
            # x_train_ = expression
            # y_train_ = status
            # x_test = expression1
            # y_test = status1
            # x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
            #                                                   test_size=1/9, random_state = random_seed, stratify = y_train_)

            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)                              

            train_dataset = CustomDataset(x_train,y_train)
            val_dataset = CustomDataset(x_val,y_val)
            test_dataset = CustomDataset(x_test,y_test) 

            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            


            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = PINNet(input_dim,pathway_info,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=20, verbose = True, path = 'checkpoint_ES.pt')

                    ##train 
                    for epoch in range(0, 200):
                        #print(epoch)
                        for batch_idx, samples in enumerate(train_loader):
                            _,_ = self.train_step(samples,training = True)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            prob, true = self.train_step(samples,training = False)

                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())

                        val_auc, _, _, _ = self.evalutaion(y_true,y_prob)
                        #print(val_auc)

                        early_stopping(val_auc, self.model, epoch)
                        if early_stopping.early_stop:
                            break

                    ##validation 
                    self.model = torch.load('checkpoint_ES.pt')
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    val_auc, val_precision, val_recall, val_f1 = self.evalutaion(y_true,y_prob)

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())
                    test_auc,_,_,_ = self.evalutaion(y_true,y_prob)
                    if test_auc > best_test_auc:
                        best_test_auc = test_auc
                        best_fold = fold+1
                        print("test auc of this fold: ", test_auc)
                        print("Best fold so far: ",best_fold)
                        ## train
                        y_prob_train, y_true_train = [],[]
                        for batch_idx, samples in enumerate(train_loader):     
                            prob, true = self.train_step(samples,training = False)
                            y_prob_train.extend(prob.detach().cpu().numpy())
                            y_true_train.extend(true.cpu().numpy())
                        ## independent test
                        y_prob_ind, y_true_ind = [], []
                        for batch_idx, samples in enumerate(test_loader1):     
                            prob, true = self.train_step(samples,training = False)
                            y_prob_ind.extend(prob.detach().cpu().numpy())
                            y_true_ind.extend(true.cpu().numpy())
                        shap_values, shap_values_feat = self.get_shap_values(x_train, expression1)
                    else:
                        print("test auc of this fold: ", test_auc)
                        print("Best fold so far: ",best_fold)
 
        return y_prob_ind, y_true_ind, y_prob_train, y_true_train, shap_values, shap_values_feat
        
    def train_step(self, batch_item, training):
        data,label = batch_item
        data = data.to(self.device)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        
    def get_shap_values(self, x_train, x_val):
            self.model.eval()
            explainer = shap.GradientExplainer((self.model, self.model.fc2), torch.tensor(x_train, dtype=torch.float32).to(self.device))
            shap_values = explainer.shap_values(torch.tensor(x_val, dtype=torch.float32).to(self.device))
            explainer_feat = shap.GradientExplainer(self.model, torch.tensor(x_train, dtype=torch.float32).to(self.device))
            shap_values_feat = explainer_feat.shap_values(torch.tensor(x_val, dtype=torch.float32).to(self.device))
            return shap_values, shap_values_feat
        
def independent_test():
    parser = argparse.ArgumentParser(description='Independent Test')
    parser.add_argument('--input_train_csv', default='gene.average.beta.by.intensity.train.csv', help='Input training csv file of X_i. Rows are genes, columns are samples.')
    parser.add_argument('--input_test_csv', default='gene.average.beta.by.intensity.test.csv', help='Input testing csv file of X_i. Rows are genes, columns are samples.')
    parser.add_argument('--input_label_train', default='label_train.csv', help='Input training label file path.')
    parser.add_argument('--input_label_test', default='label_test.csv', help='Input testing label file path.')
    parser.add_argument('--input_pathway', default='pathway_gobp.csv', help='Pathway information from dataprep.')
    parser.add_argument('--mlp', action='store_true', help='DNN only model.')

    args = parser.parse_args()

    input_file_train_csv = args.input_train_csv
    input_file_test_csv = args.input_test_csv
    input_file_label_train = args.input_label_train
    input_file_label_test = args.input_label_test
    input_file_path = args.input_pathway
    mlp = args.mlp
    
    pathway_info, expression, status, features,path_features = preprocessing1(input_file_path, input_file_train_csv, input_file_label_train)
    pathway_info1, expression1, status1, features1,path_features1 = preprocessing1(input_file_path, input_file_test_csv, input_file_label_test)
    print(path_features[0:10])
    
    trainArgs = {}
    trainArgs['pathway_info'] = pathway_info
    trainArgs['features'] = features
    trainArgs['num_fc_list'] = [64]
    trainArgs['lr_list'] = [0.0005]
    #trainArgs['num_fc_list'] = [32]
    #trainArgs['lr_list'] = [0.0001]
    trainArgs['device'] = '0'
    trainArgs['seed'] = 0
    trainArgs['filename'] = 'result.csv'
    trainArgs['expression1'] = expression1
    trainArgs['status1'] = status1
    trainArgs['expression'] = expression
    trainArgs['status'] = status
    
    
    train = train_test(trainArgs)
    y_prob, y_test, y_prob_train, y_true_train, shap_values, shap_values_feat= train.kfold()
    
    scores = []
    index = []
    
    y_true = np.array(y_test).astype(int)
    y_prob = np.array(y_prob)
    auc = roc_auc_score(y_true,y_prob)
    prauc = average_precision_score(y_true,y_prob)
    pr1, rc1, f1 = fscore(y_true,y_prob,1)
    pr2, rc2, f2 = fscore(y_true,y_prob,2)
    pr5, rc5, f5 = fscore(y_true,y_prob,5)
    pr10, rc10, f10 = fscore(y_true,y_prob,10)
    threshold = get_threshold(y_true, y_prob)
    mcc, threshold = mcc_score(y_true, y_prob,threshold)
    y_pred = np.greater(y_prob, threshold).astype(int)
    acc = accuracy_score(y_true, y_pred)
    tpr, tnr, ppv, npv = calculate_metrics(y_true, y_pred)
    scores.append([auc,prauc, mcc,threshold, acc,tpr, tnr, ppv, npv,f1,f2,f5,f10])
    index.append("Test")
    y_true = np.array(y_true_train).astype(int)
    y_prob = np.array(y_prob_train)
    auc = roc_auc_score(y_true,y_prob)
    prauc = average_precision_score(y_true,y_prob)
    pr1, rc1, f1 = fscore(y_true,y_prob,1)
    pr2, rc2, f2 = fscore(y_true,y_prob,2)
    pr5, rc5, f5 = fscore(y_true,y_prob,5)
    pr10, rc10, f10 = fscore(y_true,y_prob,10)
    threshold = get_threshold(y_true, y_prob)
    mcc, threshold = mcc_score(y_true, y_prob,threshold)
    y_pred = np.greater(y_prob, threshold).astype(int)
    acc = accuracy_score(y_true, y_pred)
    tpr, tnr, ppv, npv = calculate_metrics(y_true, y_pred)
    scores.append([auc,prauc, mcc,threshold, acc,tpr, tnr, ppv, npv,f1,f2,f5,f10])
    index.append("Train")
    

    shap = shap_values
    feature_importance = np.sum(np.abs(shap[1]), axis=0)[64:]
    print(feature_importance.shape)
    positive_sum = np.sum(shap[1] * (shap[1] > 0), axis=0)[64:]
    negative_sum = np.sum(shap[1] * (shap[1] < 0), axis=0)[64:]
    
    fi_result = pd.DataFrame({"path_name":path_features,"importance":feature_importance, "pos":positive_sum,"neg": np.abs(negative_sum)})
    
    df_sorted = fi_result.sort_values(by='importance', ascending=False)
    df_sorted.to_csv("Pathway_Importance.csv",index=False)
    shap = shap_values_feat
    feature_importance = np.sum(np.abs(shap[1]), axis=0)
    print(feature_importance.shape)
    positive_sum = np.sum(shap[1] * (shap[1] > 0), axis=0)
    negative_sum = np.sum(shap[1] * (shap[1] < 0), axis=0)
    print(len(features),len(feature_importance),len(positive_sum),len(negative_sum))
    fi_result = pd.DataFrame({"gene_name":features,"importance":feature_importance, "pos":positive_sum,"neg": np.abs(negative_sum)})
    df_sorted = fi_result.sort_values(by='importance', ascending=False)
    df_sorted.to_csv("Gene_Feature_Importance.csv",index=False)
    
    data = np.array(scores)
    col_names = ["AUROC","PRAUC", "MCC", "Threshold(by MCC)", "Accuracy","TPR(recall)", "TNR(specificity)", "PPV(precision)", "NPV","Fscore(beta=1)", "Fscore(beta=2)", "Fscore(beta=5)","Fscore(beta=10)"]
    df = pd.DataFrame(data, index=index, columns=col_names)
    
    df.to_csv("Result.csv")
    
if __name__ == '__main__':
    independent_test()


import os, copy, torch, random, time, datetime
import numpy as np
import pandas as pd
from torch import nn
import torch.nn.functional as F
from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from sklearn import metrics
from sklearn.feature_selection import mutual_info_classif
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score, average_precision_score
from sklearn.metrics import precision_recall_curve
from statistics import mean 
from imblearn.over_sampling import SMOTE
import shap
import pickle
from PROMINENT_methylation.model import *
from PROMINENT_methylation.utils import *
import matplotlib.pyplot as plt

class train_kfold:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : gene expression data
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        directory = './CV_best_model'
        if not os.path.exists(directory):
            os.makedirs(directory)

    def kfold(self):
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        input_dim = x_data.shape[1]
        y_data = trainArgs['y_data']

        pathway_info = trainArgs['pathway_info'].to(self.device)
        num_pathway = pathway_info.shape[0]
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']

        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])   

        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            
            dict_data = {"train": train_index, "test": test_index}
            with open(f"../../data/folds/fold_{fold+1}.pkl", 'wb') as file:
                pickle.dump(dict_data, file)
            test_dataset = CustomDataset(x_test,y_test)    
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    val_precision_list = []
                    val_recall_list = []
                    val_f1_list = []
                    seeds = [1,2,3,4,5,6,7,8,9,10]
                    for train_val_seed in seeds:
                        x_train, x_val, y_train, y_val, index_train, index_val = train_test_split(x_train_, y_train_, train_index, test_size=1/9, random_state = train_val_seed, stratify = y_train_)
                        smote = SMOTE(random_state=random_seed)
                        x_train, y_train = smote.fit_resample(x_train,y_train)
                        y_train = y_train.reshape(-1,1) 
                        train_dataset = CustomDataset(x_train,y_train)
                        val_dataset = CustomDataset(x_val,y_val)
                        train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
                        val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
                        self.model = PINNet(input_dim,pathway_info,num_fc)
                        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                        self.criterion = nn.CrossEntropyLoss()
                        self.model = self.model.to(self.device)
                        early_stopping = EarlyStopping(patience=20, verbose = True, path = 'checkpoint_ES.pt')

                        ##train 
                        for epoch in range(0, 200):
                            for batch_idx, samples in enumerate(train_loader):
                                _,_ = self.train_step(samples,training = True)
                            ##early stopping
                            y_prob, y_true = [],[]
                            for batch_idx, samples in enumerate(val_loader):
                                prob, true = self.train_step(samples,training = False)

                                y_prob.extend(prob.detach().cpu().numpy())
                                y_true.extend(true.cpu().numpy())

                            val_auc, _, _, _ = self.evalutaion(y_true,y_prob)
                            print(val_auc)
                            early_stopping(val_auc, self.model, epoch)
                            if early_stopping.early_stop:
                                break

                        ##validation 
                        self.model = torch.load('checkpoint_ES.pt')
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):     
                            prob, true = self.train_step(samples,training = False)
                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())

                        val_auc, val_precision, val_recall, val_f1 = self.evalutaion(y_true,y_prob)
                        val_auc_list.append(val_auc)
                        val_precision_list.append(val_precision)
                        val_recall_list.append(val_recall)
                        val_f1_list.append(val_f1)
                    mean_val_auc = np.mean(val_auc_list)
                    mean_val_precision = np.mean(val_precision_list)
                    mean_val_recall = np.mean(val_recall_list)
                    mean_val_f1 = np.mean(val_f1_list)
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1 = self.evalutaion(y_true,y_prob)
                    
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [mean_val_auc], 'Valid_Precision': [mean_val_precision], 
                                            'Valid_Recall': [mean_val_recall], 'Valid_F1': [mean_val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1]})], ignore_index=True)
        
        return result
        
    def train_step(self, batch_item, training):
        data,label = batch_item
        data = data.to(self.device)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        

class train_kfold1:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : gene expression data
            y_data          : patient label
            pathway_info    : pathway matrics
            features        : feature names list
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        directory = './CV_best_model'
        if not os.path.exists(directory):
            os.makedirs(directory)

    def kfold(self):
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        input_dim = x_data.shape[1]
        y_data = trainArgs['y_data']

        pathway_info = trainArgs['pathway_info'].to(self.device)
        num_pathway = pathway_info.shape[0]
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']

        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])

        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            
            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)                              
            print(torch.tensor(x_train).shape)
            
            train_dataset = CustomDataset(x_train,y_train)
            val_dataset = CustomDataset(x_val,y_val)
            test_dataset = CustomDataset(x_test,y_test) 
                       
            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            
                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = PINNet(input_dim,pathway_info,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=10, verbose = True, path = 'checkpoint_ES.pt')
                   
                    ##train 
                    for epoch in range(0, 200):
                        for batch_idx, samples in enumerate(train_loader):
                            _,_ = self.train_step(samples,training = True)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            prob, true = self.train_step(samples,training = False)

                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())
                        val_auc, _, _, _ = self.evalutaion(y_true,y_prob)

                        early_stopping(val_auc, self.model, epoch)
                        
                        if early_stopping.early_stop:
                            break
                            
                    ##validation 
                    self.model = torch.load('checkpoint_ES.pt')
                    
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())
                    
                    val_auc, val_precision, val_recall, val_f1 = self.evalutaion(y_true,y_prob)
                    print(">>[val] auc : {:.4f}, precision : {:.4f}, recall : {:.4f}, f1 : {:.4f}"
                          .format(val_auc, val_precision, val_recall, val_f1))                    
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 
                    torch.save({'model': self.model.state_dict(), 'fold': fold,'learning_rate':lr,'num_fc':num_fc},
                               './CV_best_model/best_model_fold_{}.pt'.format(fold))
                    
                    ## Get SHAP values
                    features = trainArgs['features']
                    shap_values = self.get_shap_values(x_train, x_val)
                    feature_importance = pd.DataFrame({"Feature": features, f'Importance{fold+1}': np.mean(np.abs(shap_values), axis=0)})
                    top20 = feature_importance.sort_values(by=f'Importance{fold+1}').head(20)
                    plt.barh(top20['Feature'], top20[f'Importance{fold+1}'])
                    plt.title(f'Top 20 Features by Importance of Fold {fold+1}')
                    plt.xlabel('Importance')
                    plt.ylabel('Feature')
                    plt.show()
                    if fold == 0:
                        fi_result = feature_importance
                    else:
                        fi_result = pd.concat([fi_result, feature_importance.iloc[:,1]],axis=1)
                    
                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1 = self.evalutaion(y_true,y_prob)

                    print(">>[test] auc : {:.4f}, precision : {:.4f}, recall : {:.4f}, f1 : {:.4f}\n"
                          .format(test_auc, test_precision, test_recall, test_f1))
                    
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [val_auc], 'Valid_Precision': [val_precision], 
                                            'Valid_Recall': [val_recall], 'Valid_F1': [val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1]})], ignore_index=True)
        
        return result, fi_result
        
    def train_step(self, batch_item, training):
        data,label = batch_item
        data = data.to(self.device)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
                
            loss.backward()
            self.optimizer.step()
            e = shap.DeepExplainer(self.model, data)
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]  
            return prob, true
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        
    def get_shap_values(self, x_train, x_val):
        self.model.eval()
        explainer = shap.GradientExplainer(self.model, torch.tensor(x_train, dtype=torch.float32).to(self.device))
        shap_values = explainer.shap_values(torch.tensor(x_val, dtype=torch.float32).to(self.device))[1]
        return np.array(shap_values)
    
class train_kfold2:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : gene expression data
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        print(self.device)

    def kfold(self):
        print(self.device)
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        input_dim = x_data.shape[1]
        y_data = trainArgs['y_data']

        pathway_info = trainArgs['pathway_info'].to(self.device)
        num_pathway = pathway_info.shape[0]
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']

        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])   
        shap_ls = []
        shap_feat_ls = []
        x_tests = []
        test_preds = []
        preds_fn = trainArgs['filename']
        shap_fn = trainArgs['filename2']
        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            dict_data = {"train": train_index, "test": test_index}
            
            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)  
            
            train_dataset = CustomDataset(x_train,y_train)
            val_dataset = CustomDataset(x_val,y_val)
            test_dataset = CustomDataset(x_test,y_test)  
                       
            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            
                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = PINNet(input_dim,pathway_info,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=10, verbose = True, path = f'checkpoint_ES.pt')
                   
                    ##train 
                    for epoch in range(0, 200):
                        for batch_idx, samples in enumerate(train_loader):
                            _,_ = self.train_step(samples,training = True)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            prob, true = self.train_step(samples,training = False)

                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())

                        val_auc, _, _, _, _ = self.evalutaion(y_true,y_prob)

                        early_stopping(val_auc, self.model, epoch)
                        if early_stopping.early_stop:
                            break
                            
                    ##validation 
                    self.model = torch.load(f'checkpoint_ES.pt')
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    val_auc, val_precision, val_recall, val_f1, _ = self.evalutaion(y_true,y_prob)
                 
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1, test_pr_auc = self.evalutaion(y_true,y_prob)
                    test_preds.append([y_true,y_prob])
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [val_auc], 'Valid_Precision': [val_precision], 
                                            'Valid_Recall': [val_recall], 'Valid_F1': [val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1], 'Test_PrAUC': [test_pr_auc]})], ignore_index=True)
                    
                    ##SHAP
                    shap_values, shap_values_feat = self.get_shap_values(x_train, x_test)
                    shap_ls.append(shap_values)
                    shap_feat_ls.append(shap_values_feat)
                    x_tests.append(x_test)
        with open(shap_fn, 'wb') as file:
            pickle.dump({"path":shap_ls,"feat":shap_feat_ls}, file)
        with open(preds_fn, 'wb') as file:
            pickle.dump(test_preds, file)
        return result
        
    def train_step(self, batch_item, training):
        data,label = batch_item
        data = data.to(self.device)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        pr_auc = average_precision_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1, pr_auc

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        
    def get_shap_values(self, x_train, x_val):
            self.model.eval()
            explainer = shap.GradientExplainer((self.model, self.model.fc2), torch.tensor(x_train, dtype=torch.float32).to(self.device))
            shap_values = explainer.shap_values(torch.tensor(x_val, dtype=torch.float32).to(self.device))
            explainer_feat = shap.GradientExplainer(self.model, torch.tensor(x_train, dtype=torch.float32).to(self.device))
            shap_values_feat = explainer_feat.shap_values(torch.tensor(x_val, dtype=torch.float32).to(self.device))
            return shap_values, shap_values_feat

class train_kfold_mlp:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : gene expression data
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        directory = './CV_best_model'
        if not os.path.exists(directory):
            os.makedirs(directory)

    def kfold(self):
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        input_dim = x_data.shape[1]
        y_data = trainArgs['y_data']
        pathway_info = trainArgs['pathway_info'].to(self.device)
        num_pathway = pathway_info.shape[0]
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']

        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])   
        
        test_preds = []
        preds_fn = trainArgs['filename']
        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            
            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)                              
            
 
            train_dataset = CustomDataset(x_train,y_train)
            val_dataset = CustomDataset(x_val,y_val)
            test_dataset = CustomDataset(x_test,y_test)
                       
            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            
                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = MLP(input_dim,pathway_info,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=20, verbose = True, path = f'checkpoint_ES.pt')
                   
                    ##train 
                    for epoch in range(0, 200):
                        for batch_idx, samples in enumerate(train_loader):
                            _,_ = self.train_step(samples,training = True)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            prob, true = self.train_step(samples,training = False)

                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())

                        val_auc, _, _, _, _ = self.evalutaion(y_true,y_prob)

                        early_stopping(val_auc, self.model, epoch)
                        if early_stopping.early_stop:
                            break
                            
                    ##validation 
                    self.model = torch.load(f'checkpoint_ES.pt')
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    val_auc, val_precision, val_recall, val_f1, _ = self.evalutaion(y_true,y_prob)
                 
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1, test_pr_auc = self.evalutaion(y_true,y_prob)
                    test_preds.append([y_true,y_prob])
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [val_auc], 'Valid_Precision': [val_precision], 
                                            'Valid_Recall': [val_recall], 'Valid_F1': [val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1], 'Test_PrAUC': [test_pr_auc]})], ignore_index=True)
        with open(preds_fn, 'wb') as file:
            pickle.dump(test_preds, file)
        return result
    
    def train_step(self, batch_item, training):
        data,label = batch_item
        data = data.to(self.device)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(data)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        pr_auc = average_precision_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1, pr_auc

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
               
class train_kfold3:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : Matrix data with gene features and methylation
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    def kfold(self):
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        y_data = trainArgs['y_data']
        pathway_info = trainArgs['pathway_info']
        sel_feat_num = trainArgs['sel_feat_num']
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']

        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])   
        shap_ls = []
        x_tests = []
        sel_feat_idx = {}
        test_preds = []
        preds_fn = trainArgs['filename']
        feat_idx_fn = trainArgs['sel_feat_fn']
        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            
            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)                              

            if os.path.exists(feat_idx_fn):
                with open(feat_idx_fn, 'rb') as file:
                    sel_feat_idx = pickle.load(file)
                    kept_indice1 = sel_feat_idx[f'fold{fold}'][0]
                    kept_indice2 = sel_feat_idx[f'fold{fold}'][1]
            else:
                print("Run feature selection first!")
            
            pathway_info_input = pathway_info[:,kept_indice2]
            pathway_info_input = pathway_info_input.to(self.device)
            input_size_gene = len(kept_indice2)
            input_size_meth = sel_feat_num
            
            train_dataset = CustomDataset(x_train[:,kept_indice1],y_train)
            val_dataset = CustomDataset(x_val[:,kept_indice1],y_val)
            test_dataset = CustomDataset(x_test[:,kept_indice1],y_test) 
                       
            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            
                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = PINNet3(input_size_gene,sel_feat_num,pathway_info_input,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=10, verbose = True, path = f'/ix/hpark/laz64/ipf/dlmethylation/run{random_seed}/checkpoint_ES.pt')
                   
                    ##train 
                    for epoch in range(0, 200):
                        for batch_idx, samples in enumerate(train_loader):
                            #print(samples[0].shape)
                            _,_ = self.train_step(samples,training = True, num_gene = input_size_gene)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)
                            #print(true, prob)
                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())
                        val_auc, _, _, _, _ = self.evalutaion(y_true,y_prob)

                        early_stopping(val_auc, self.model, epoch)
                        if early_stopping.early_stop:
                            break
                            
                    ##validation 
                    self.model = torch.load(f'/ix/hpark/laz64/ipf/dlmethylation/run{random_seed}/checkpoint_ES.pt')
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    val_auc, val_precision, val_recall, val_f1, _ = self.evalutaion(y_true,y_prob)
                 
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1, test_pr_auc = self.evalutaion(y_true,y_prob)
                    test_preds.append([y_true,y_prob])
                    
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [val_auc], 'Valid_Precision': [val_precision], 
                                            'Valid_Recall': [val_recall], 'Valid_F1': [val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1], 'Test_PrAUC': [test_pr_auc]})], ignore_index=True)
                    
                    ##SHAP
                    # shap_values = self.get_shap_values(x_train[:,kept_indice], x_test[:,kept_indice],num_gene=input_size_gene)
                    # shap_ls.append(shap_values)
                    # x_tests.append(x_test)
        # with open("shap.pkl", 'wb') as file:
        #     pickle.dump(shap_ls, file)
        with open(preds_fn, 'wb') as file:
            pickle.dump(test_preds, file)
        return result
        
    def train_step(self, batch_item, training,num_gene):
        data,label = batch_item
        data = data.to(self.device)
        input_gene = data[:,:num_gene]
        #print(input_gene.shape)
        input_meth = data[:,num_gene:]
        #print(input_meth.shape)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(input_gene,input_meth)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                print(input_gene)
                print(input_meth)
                out = self.model(input_gene,input_meth)
                print(out)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        pr_auc = average_precision_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1, pr_auc

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        
    def get_shap_values(self, x_train, x_val, num_gene):
            self.model.eval()
            input_gene = torch.tensor(x_train[:,:num_gene], dtype=torch.float32).to(self.device)
            input_meth = torch.tensor(x_train[:,num_gene:], dtype=torch.float32).to(self.device)
            input_gene_val = torch.tensor(x_val[:,:num_gene], dtype=torch.float32).to(self.device)
            input_meth_val = torch.tensor(x_val[:,num_gene:], dtype=torch.float32).to(self.device)
            explainer = shap.GradientExplainer((self.model, self.model.fc2), [input_gene,input_meth])
            shap_values = explainer.shap_values([input_gene_val, input_meth_val])
            return shap_values
        
class train_kfold4:
    def __init__(self, trainArgs):
   
        """
        args:
            x_data          : Matrix data with gene features and methylation
            y_data          : patient label
            pathway_info    : pathway matrics
            num_fc_list     : number of fully connected nodes 
            lr_list         : learning rate
            device          : GPU device
        Returns:
            AUC, Precision, Recall, F1
        """
        self.trainArgs = trainArgs
        self.seed_worker(trainArgs['seed'])
        os.environ["CUDA_VISIBLE_DEVICES"]=trainArgs['device']
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        directory = './CV_best_model'
        if not os.path.exists(directory):
            os.makedirs(directory)

    def kfold(self):
        trainArgs = self.trainArgs
        x_data = trainArgs['x_data']
        y_data = trainArgs['y_data']
        pathway_info = trainArgs['pathway_info']
        sel_feat_num = trainArgs['sel_feat_num']
        num_fc_list = trainArgs['num_fc_list']
        lr_list = trainArgs['lr_list']
        random_seed = trainArgs['seed']
        input_size_gene = pathway_info.shape[1]
        input_size_meth = sel_feat_num
        pathway_info_input = pathway_info.to(self.device)
        kfold = StratifiedKFold(n_splits = 5, shuffle=True, random_state = random_seed)
        
        result = pd.DataFrame(columns=['hyperparam','Fold', 'Valid_AUC','Valid_Precision','Valid_Recall','Valid_F1',
                                       'Test_AUC','Test_Precision','Test_Recall','Test_F1'])   
        shap_ls = []
        x_tests = []
        sel_feat_idx = {}
        test_preds = []
        for fold, (train_index, test_index) in enumerate(kfold.split(x_data, y_data)):   
            print('****************************************************************************')
            print('Fold {} / {}'.format(fold + 1 , kfold.get_n_splits()))
            print('****************************************************************************')
            x_train_ = x_data[train_index]
            y_train_ = y_data[train_index] 
            x_test = x_data[test_index]  
            y_test = y_data[test_index] 
            x_train, x_val, y_train, y_val = train_test_split(x_train_, y_train_, 
                                                              test_size=1/9, random_state = random_seed, stratify = y_train_)
            
            smote = SMOTE(random_state=random_seed)
            x_train, y_train = smote.fit_resample(x_train,y_train)
            y_train = y_train.reshape(-1,1)                              

            if os.path.exists("mi_feat_idx_gobp.pkl"):
                with open("mi_feat_idx_gobp.pkl", 'rb') as file:
                    sel_feat_idx = pickle.load(file)
                    kept_indice3 = sel_feat_idx[f'fold{fold}'][2]
            else:
                print("Run feature selection first!")
            
            kept_indice1 = [i for i in range(input_size_gene)] + [x + input_size_gene for x in kept_indice3]
            print(len(kept_indice1),input_size_gene)
            train_dataset = CustomDataset(x_train[:,kept_indice1],y_train)
            val_dataset = CustomDataset(x_val[:,kept_indice1],y_val)
            test_dataset = CustomDataset(x_test[:,kept_indice1],y_test) 
                       
            train_loader = DataLoader(dataset = train_dataset, batch_size = 64, shuffle = True)
            val_loader = DataLoader(dataset = val_dataset, batch_size = 64, shuffle = False)
            test_loader = DataLoader(dataset = test_dataset, batch_size = 64, shuffle = False)            
                   
            best_val_auc = 0
            for  lr in lr_list:
                for num_fc in num_fc_list:
                    val_auc_list = []
                    test_auc_list = []
                    self.model = PINNet3(input_size_gene,sel_feat_num,pathway_info_input,num_fc)
                    self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr, weight_decay = 0)
                    self.criterion = nn.CrossEntropyLoss()
                    self.model = self.model.to(self.device)
                    early_stopping = EarlyStopping(patience=10, verbose = True, path = 'checkpoint_ES.pt')
                   
                    ##train 
                    for epoch in range(0, 200):
                        for batch_idx, samples in enumerate(train_loader):
                            #print(samples[0].shape)
                            _,_ = self.train_step(samples,training = True, num_gene = input_size_gene)
                        ##early stopping
                        y_prob, y_true = [],[]
                        for batch_idx, samples in enumerate(val_loader):
                            #print(samples)
                            
                            prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)
                            #print(true, prob)
                            y_prob.extend(prob.detach().cpu().numpy())
                            y_true.extend(true.cpu().numpy())
                        #print(y_true, y_prob)
                        val_auc, _, _, _, _ = self.evalutaion(y_true,y_prob)

                        early_stopping(val_auc, self.model, epoch)
                        if early_stopping.early_stop:
                            break
                            
                    ##validation 
                    self.model = torch.load('checkpoint_ES.pt')
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(val_loader):     
                        prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)
                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    val_auc, val_precision, val_recall, val_f1, _ = self.evalutaion(y_true,y_prob)
                 
                    
                    if val_auc > best_val_auc:
                        best_val_auc = val_auc 

                    ##test
                    y_prob, y_true = [],[]
                    for batch_idx, samples in enumerate(test_loader):     
                        prob, true = self.train_step(samples,training = False, num_gene = input_size_gene)

                        y_prob.extend(prob.detach().cpu().numpy())
                        y_true.extend(true.cpu().numpy())

                    test_auc, test_precision, test_recall, test_f1, test_pr_auc = self.evalutaion(y_true,y_prob)
                    
                    test_preds.append([y_true,y_prob])
                    result = pd.concat([result, pd.DataFrame({'hyperparam': ["lr:{} / num_fc:{}".format(str(lr),str(num_fc))],'Fold':[fold],
                                            'Valid_AUC': [val_auc], 'Valid_Precision': [val_precision], 
                                            'Valid_Recall': [val_recall], 'Valid_F1': [val_f1],
                                            'Test_AUC': [test_auc], 'Test_Precision': [test_precision], 
                                            'Test_Recall': [test_recall], 'Test_F1': [test_f1], 'Test_PrAUC': [test_pr_auc]})], ignore_index=True)
                    
                    ##SHAP
                    # shap_values = self.get_shap_values(x_train[:,kept_indice], x_test[:,kept_indice],num_gene=input_size_gene)
                    # shap_ls.append(shap_values)
                    # x_tests.append(x_test)
        # with open("shap.pkl", 'wb') as file:
        #     pickle.dump(shap_ls, file)
        with open("test_preds.pkl", 'wb') as file:
            pickle.dump(test_preds, file)
        return result
        
    def train_step(self, batch_item, training,num_gene):
        data,label = batch_item
        data = data.to(self.device)
        input_gene = data[:,:num_gene]
        #print(input_gene.shape)
        input_meth = data[:,num_gene:]
        #print(input_meth.shape)
        label = label.to(self.device)
        if training is True:
            self.model.train()
            self.optimizer.zero_grad()
            with torch.cuda.amp.autocast():
                out = self.model(input_gene,input_meth)
                true = torch.reshape(label,(-1,))
                loss = self.criterion(out,true)
                prob = out[:,1]                
            loss.backward()
            self.optimizer.step()
            return prob, true
        else:
            self.model.eval()
            with torch.no_grad():
                out = self.model(input_gene,input_meth)
                true = torch.reshape(label,(-1,))
                prob = out[:,1]   
            return prob, true  
        
    def evalutaion(self, y_true, y_prob):
        np.seterr(divide='ignore', invalid='ignore')
        auc = roc_auc_score(y_true,y_prob)
        pr_auc = average_precision_score(y_true,y_prob)
        precision,recall,_ = precision_recall_curve(y_true,y_prob)
        f1 = (2*precision*recall)/(precision+recall)
        idx = np.nanargmax(f1)
        pr = precision[idx] 
        rc = recall[idx] 
        f1 = f1[idx] 
        return auc, pr, rc, f1, pr_auc

    def seed_worker(self, random_seed):
        torch.manual_seed(random_seed)
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)  # if use multi-GPU
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        np.random.seed(random_seed)
        random.seed(random_seed)
        
    def get_shap_values(self, x_train, x_val, num_gene):
            self.model.eval()
            input_gene = torch.tensor(x_train[:,:num_gene], dtype=torch.float32).to(self.device)
            input_meth = torch.tensor(x_train[:,num_gene:], dtype=torch.float32).to(self.device)
            input_gene_val = torch.tensor(x_val[:,:num_gene], dtype=torch.float32).to(self.device)
            input_meth_val = torch.tensor(x_val[:,num_gene:], dtype=torch.float32).to(self.device)
            explainer = shap.GradientExplainer((self.model, self.model.fc2), [input_gene,input_meth])
            shap_values = explainer.shap_values([input_gene_val, input_meth_val])
            return shap_values


from setuptools import setup

setup(
    name='PROMINENT_methylation',
    version='0.1.10',
    packages=['PROMINENT_methylation'],
    description='PROMINENT: AN INTERPRETABLE DEEP LEARNING METHOD TO PREDICT PHENOTYPES USING DNA METHYLATION',
    url='https://github.com/cloudmacchiato/dlmethylation',
    author='Laizhi Zhang',
    author_email='laz64@pitt.edu',
    license='MIT',
    entry_points={
        'console_scripts': [
            'PROMINENT-data_prepare = PROMINENT_methylation.dataprep:dataprep',
            'PROMINENT-train_test_cv = PROMINENT_methylation.train_test:train',
            'PROMINENT-scores = PROMINENT_methylation.scores:get_scores',
            'PROMINENT-model_interpret = PROMINENT_methylation.interpret:get_feature_importance',
            'PROMINENT-model_train_test_independent = PROMINENT_methylation.independent_test:independent_test'
        ]
    },
    install_requires=['numpy',
                      'shap',
                      'seaborn',
                      'matplotlib',
                      'torch',
                      'imblearn'
    ]
)



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import shap
import pickle
import seaborn as sns
from sklearn.metrics import roc_curve, precision_recall_curve, auc
from sklearn.metrics import roc_auc_score, average_precision_score
from sklearn.metrics import precision_recall_curve, matthews_corrcoef,accuracy_score
from scipy.stats import shapiro,mannwhitneyu,ttest_ind
import seaborn as sns
import argparse

def softmax(x):
    # Subtract the maximum value from each element for numerical stability
    x_exp = np.exp(x - np.max(x, axis=1, keepdims=True))
    # Calculate the softmax probabilities
    return x_exp / np.sum(x_exp, axis=1, keepdims=True)
def fscore(true, prob, beta):
    precision,recall,_ = precision_recall_curve(true,prob)
    f = (1+beta**2)*(precision*recall)/(beta**2*precision+recall)
    idx = np.nanargmax(f)
    pr = precision[idx] 
    rc = recall[idx] 
    f = f[idx]
    return pr, rc, f
def calculate_metrics(y_true, y_pred):
    y_true = np.array(y_true).astype(int)
    TP = np.sum((y_pred == 1) & (y_true == 1))
    TN = np.sum((y_pred == 0) & (y_true == 0))
    FP = np.sum((y_pred == 1) & (y_true == 0))
    FN = np.sum((y_pred == 0) & (y_true == 1))
    TPR = TP / (TP + FN)
    TNR = TN / (TN + FP)
    PPV = TP / (TP + FP)
    NPV = TN / (TN + FN)
    return TPR, TNR, PPV, NPV
def mcc_score(y_true, y_prob):
    mccs = []
    thresholds = np.arange(0,1,0.001)
    for threshold in thresholds:
        y_pred = np.greater(y_prob, threshold).astype(int)
        mcc = matthews_corrcoef(y_true, y_pred)
        mccs.append(mcc)
    mccs = np.array(mccs)
    max_mcc = mccs.max() 
    max_mcc_threshold =  thresholds[mccs.argmax()]
    return max_mcc, max_mcc_threshold

def get_scores():
    parser = argparse.ArgumentParser(description='Caculate classification scores.')
    parser.add_argument('--input_pkl', default='pred.pkl', help='Precition output file from training. Default: pred.pkl')
    parser.add_argument('--output', default='scores.csv', help='Output file path. csv file.')

    args = parser.parse_args()

    input_file_pkl = args.input_pkl
    output_file = args.output
    
    with open(input_file_pkl, 'rb') as file:
        test_preds = pickle.load(file)
    scores = []
    for i in range(0,5):
        y_true = np.array(test_preds[i][0]).astype(int)
        y_prob = np.array(test_preds[i][1])
        auc = roc_auc_score(y_true,y_prob)
        prauc = average_precision_score(y_true,y_prob)
        pr1, rc1, f1 = fscore(y_true,y_prob,1)
        pr2, rc2, f2 = fscore(y_true,y_prob,2)
        pr5, rc5, f5 = fscore(y_true,y_prob,5)
        pr10, rc10, f10 = fscore(y_true,y_prob,10)
        mcc, threshold = mcc_score(y_true, y_prob)
        y_pred = np.greater(y_prob, threshold).astype(int)
        acc = accuracy_score(y_true, y_pred)
        tpr, tnr, ppv, npv = calculate_metrics(y_true, y_pred)
        scores.append([auc,prauc, mcc,threshold, acc,tpr, tnr, ppv, npv,f1,f2,f5,f10])
    data = np.array(scores)
    row_names = ['Fold1', 'Fold2', 'Fold3', 'Fold4', 'Fold5']
    col_names = ["AUROC","PRAUC", "MCC", "Threshold(by MCC)", "Accuracy","TPR(recall)", "TNR(specificity)", "PPV(precision)", "NPV","Fscore(beta=1)", "Fscore(beta=2)", "Fscore(beta=5)","Fscore(beta=10)"]
    df = pd.DataFrame(data, index=row_names, columns=col_names)
    
    df.to_csv(output_file)
    
if __name__ == '__main__':
    get_scores()
    